######################################################################################
#
#  MFRC522FF32.py
#  Python program for communication with MFRC522 NXP Mifare reader
#  through a FF32 USB->SPI converter (www.flyfish-tech.com)
#  By   : Paul Merkx
#  Date : 07-05-2014
#  Rev  : 0.1
#  
#  uses: pyff32.py   FF32 communication library
#
#  Based on:
#  MFRC522.c - Library to use ARDUINO RFID MODULE KIT 13.56 MHZ WITH TAGS SPI W AND R BY COOQROBOT.
#  Based on code Dr.Leong   ( WWW.B2CQSHOP.COM )
#  Created by Miguel Balboa (circuitito.com), Jan, 2012.
#  Rewritten by Soren Thing Andersen (access.thing.dk), fall of 2013 (Translation to English, factored, comments, anti collision, cascade levels.)
#  Rewritten by Paul Merkx, spring 2014 (Converted to python, changed to use FF32 USB to SPI converter for SPI communication.)
#
#  Code has been tested on a Raspberry Pi --[USB]-->FF32--[SPI]-->RC522
#  Note that various RC522 boards are available. Although pins are labelled SS/SCK/MOSI/MISO the board may be hardwired to use these pins for I2C or UART communication.
#  Best check voltages on pin #1 (0 volt) and pin #32 (3.3 volt) to be sure it is using SPI
#  
#  Please read this file for a technical overview and inline comments for the specific functions.
#  Search for "mf-rc522" on ebay.com to purchase the MF-RC522 board. 
#  See www.flyfish-tech.com for availability of the FF32 interface chip.
#  There are four hardware components involved:
#  1) The Raspberry PI
#  2) The FF32 USB to SPI converter PIC
#  3) The PCD (short for Proximity Coupling Device): NXP MFRC522 Contact-less Reader IC
#  4) The PICC (short for Proximity Integrated Circuit Card): A card or tag using the ISO 14443A interface, eg Mifare or NTAG203.
#  
#  The micro controller and card reader uses SPI for communication.
#  The protocol is described in the MFRC522 datasheet: http://www.nxp.com/documents/data_sheet/MFRC522.pdf
#  
#  The card reader and the tags communicate using a 13.56MHz electromagnetic field.
#  The protocol is defined in ISO/IEC 14443-3 Identification cards -- Contactless integrated circuit cards -- Proximity cards -- Part 3: Initialization and anti-collision".
#  A free version of the final draft can be found at http://wg8.de/wg8n1496_17n3613_Ballot_FCD14443-3.pdf
#  Details are found in chapter 6, Type A: Initialization and anticollision.
#  
#  If only the PICC UID is wanted, the above documents has all the needed information.
#  To read and write from MIFARE PICCs, the MIFARE protocol is used after the PICC has been selected.
#  The MIFARE Classic chips and protocol is described in the datasheets:
# 		1K:   http://www.nxp.com/documents/data_sheet/MF1S503x.pdf
#  		4K:   http://www.nxp.com/documents/data_sheet/MF1S703x.pdf
#  		Mini: http://www.idcardmarket.com/download/mifare_S20_datasheet.pdf
#  The MIFARE Ultralight chip and protocol is described in the datasheets:
# 		Ultralight:   http://www.nxp.com/documents/data_sheet/MF0ICU1.pdf
#  		Ultralight C: http://www.nxp.com/documents/short_data_sheet/MF0ICU2_SDS.pdf
#  
#  MIFARE Classic 1K (MF1S503x):
#  		Has 16 sectors * 4 blocks/sector * 16 bytes/block = 1024 bytes.
#  		The blocks are numbered 0-63.
#  		Block 3 in each sector is the Sector Trailer. See http://www.nxp.com/documents/data_sheet/MF1S503x.pdf sections 8.6 and 8.7:
#  				Bytes 0-5:   Key A
#  				Bytes 6-8:   Access Bits
#  				Bytes 9:     User data
#  				Bytes 10-15: Key B (or user data)
#  		Block 0 is read only manufacturer data.
#  		To access a block, an authentication using a key from the block's sector must be performed first.
#  		Example: To read from block 10, first authenticate using a key from sector 3 (blocks 8-11).
#  		All keys are set to FFFFFFFFFFFFh at chip delivery.
#  		Warning: Please read section 8.7 "Memory Access". It includes this text: if the PICC detects a format violation the whole sector is irreversibly blocked.
# 		To use a block in "value block" mode (for Increment/Decrement operations) you need to change the sector trailer. Use PICC_SetAccessBits() to calculate the bit patterns.
#  MIFARE Classic 4K (MF1S703x):
#  		Has (32 sectors * 4 blocks/sector + 8 sectors * 16 blocks/sector) * 16 bytes/block = 4096 bytes.
#  		The blocks are numbered 0-255.
#  		The last block in each sector is the Sector Trailer like above.
#  MIFARE Classic Mini (MF1 IC S20):
#  		Has 5 sectors * 4 blocks/sector * 16 bytes/block = 320 bytes.
#  		The blocks are numbered 0-19.
#  		The last block in each sector is the Sector Trailer like above.
#  
#  MIFARE Ultralight (MF0ICU1):
#  		Has 16 pages of 4 bytes = 64 bytes.
#  		Pages 0 + 1 is used for the 7-byte UID.
#  		Page 2 contains the last chech digit for the UID, one byte manufacturer internal data, and the lock bytes (see http://www.nxp.com/documents/data_sheet/MF0ICU1.pdf section 8.5.2)
#  		Page 3 is OTP, One Time Programmable bits. Once set to 1 they cannot revert to 0.
#  		Pages 4-15 are read/write unless blocked by the lock bytes in page 2. 
#  MIFARE Ultralight C (MF0ICU2):
#  		Has 48 pages of 4 bytes = 64 bytes.
#  		Pages 0 + 1 is used for the 7-byte UID.
#  		Page 2 contains the last chech digit for the UID, one byte manufacturer internal data, and the lock bytes (see http://www.nxp.com/documents/data_sheet/MF0ICU1.pdf section 8.5.2)
#  		Page 3 is OTP, One Time Programmable bits. Once set to 1 they cannot revert to 0.
#  		Pages 4-39 are read/write unless blocked by the lock bytes in page 2. 
#  		Page 40 Lock bytes
#  		Page 41 16 bit one way counter
#  		Pages 42-43 Authentication configuration
#  		Pages 44-47 Authentication key 
#
#######################################################################################

import pyff32
import signal
import time
  
class MFRC522:
  NRSTPD = 22
  
  MAX_LEN = 16

  PCD_IDLE       = 0x00  # no action, cancels current command execution
  PCD_MEM        = 0x01  # stores 25 bytes into the internal buffer
  PCD_GEN_RND_ID = 0x02  # generates a 10-byte random ID number
  PCD_CALC_CRC   = 0x03  # activates the CRC coprocessor or performs a self test
  PCD_TRANSMIT   = 0x04  # transmits data from the FIFO buffer
  PCD_NO_CMD	 = 0x07  # no command change, can be used to modify the CommandReg register bits without affecting the command, for example, the PowerDown bit
  PCD_RECEIVE    = 0x08  # activates the receiver circuits
  PCD_TRANSCEIVE = 0x0C  # transmits data from FIFO buffer to antenna and automatically activates the receiver after transmission
  PCD_AUTHENT    = 0x0E  # performs the MIFARE standard authentication as a reader
  PCD_RESETPHASE = 0x0F  # performs soft reset of RC522

  
  PICC_REQIDL    = 0x26
  PICC_REQALL    = 0x52
  PICC_ANTICOLL  = 0x93
  PICC_SElECTTAG = 0x93
  PICC_AUTHENT1A = 0x60
  PICC_AUTHENT1B = 0x61
  PICC_READ      = 0x30
  PICC_WRITE     = 0xA0
  PICC_DECREMENT = 0xC0
  PICC_INCREMENT = 0xC1
  PICC_RESTORE   = 0xC2
  PICC_TRANSFER  = 0xB0
  PICC_HALT      = 0x50
  
  MI_OK       = 0
  MI_NOTAGERR = 1
  MI_ERR      = 2
  
  Reserved00     = 0x00
  CommandReg     = 0x01
  CommIEnReg     = 0x02
  DivlEnReg      = 0x03
  CommIrqReg     = 0x04
  DivIrqReg      = 0x05
  ErrorReg       = 0x06
  Status1Reg     = 0x07
  Status2Reg     = 0x08
  FIFODataReg    = 0x09
  FIFOLevelReg   = 0x0A
  WaterLevelReg  = 0x0B
  ControlReg     = 0x0C
  BitFramingReg  = 0x0D
  CollReg        = 0x0E
  Reserved01     = 0x0F
  
  Reserved10     = 0x10
  ModeReg        = 0x11
  TxModeReg      = 0x12
  RxModeReg      = 0x13
  TxControlReg   = 0x14
  TxASKReg       = 0x15
  TxSelReg       = 0x16
  RxSelReg       = 0x17
  RxThresholdReg = 0x18
  DemodReg       = 0x19
  Reserved11     = 0x1A
  Reserved12     = 0x1B
  MifareReg      = 0x1C
  Reserved13     = 0x1D
  Reserved14     = 0x1E
  SerialSpeedReg = 0x1F
  
  Reserved20        = 0x20  
  CRCResultRegM     = 0x21
  CRCResultRegL     = 0x22
  Reserved21        = 0x23
  ModWidthReg       = 0x24
  Reserved22        = 0x25
  RFCfgReg          = 0x26
  GsNReg            = 0x27
  CWGsPReg          = 0x28
  ModGsPReg         = 0x29
  TModeReg          = 0x2A
  TPrescalerReg     = 0x2B
  TReloadRegH       = 0x2C
  TReloadRegL       = 0x2D
  TCounterValueRegH = 0x2E
  TCounterValueRegL = 0x2F
  
  Reserved30      = 0x30
  TestSel1Reg     = 0x31
  TestSel2Reg     = 0x32
  TestPinEnReg    = 0x33
  TestPinValueReg = 0x34
  TestBusReg      = 0x35
  AutoTestReg     = 0x36
  VersionReg      = 0x37
  AnalogTestReg   = 0x38
  TestDAC1Reg     = 0x39
  TestDAC2Reg     = 0x3A
  TestADCReg      = 0x3B
  Reserved31      = 0x3C
  Reserved32      = 0x3D
  Reserved33      = 0x3E
  Reserved34      = 0x3F
    
  serNum = []
  
  def __init__(self):
    self.FF32_Init()
    self.MFRC522_Init()
    
  def MFRC522_Reset(self):
    self.Write_MFRC522(self.CommandReg, self.PCD_RESETPHASE)
  
  def Write_MFRC522(self,addr,val):
    # Write to MFRC522. 
	# Address byte: B7=0 for write, B6..B1=register address, B0=0 (see par 8.1.2.3 datasheet)
    # val&0xFF to convert int to hex (don't know other way to do that)
    bytes = bytearray([(addr<<1)&0x7E,val&0xFF])
    with pyff32.FF32() as ff32:
        try:
            ff32.writeSPIBus(bytes)
#            print("Write register-value: [{0:02X}-{1:02X}]".format(addr,val))
        except Exception as ex:
            print("Write_MFRC522: *** Error (WriteSPIBus): " + ex.message)

  
  def Read_MFRC522(self,addr):
    # Read 1 byte from MFRC522 register passed as address. 
	# Address byte: B7=1 for read, B6..B1=address, B0=0 (see par 8.1.2.3 datasheet)
    bytes = bytearray([(addr<<1)&0x7E | 0x80])
    with pyff32.FF32() as ff32:
        try:
            bytes = ff32.readSPIBus(1, bytes)
#            print("Read register-value: [{0:02X}-{1:02X}]".format(addr,(bytes[0])))
        except Exception as ex:
            print("Read_MFRC522: *** Error (ReadSPIBus): " + ex.message)
        
    return bytes[0]
  
  def SetBitMask(self, reg, mask):
    tmp = self.Read_MFRC522(reg)
    self.Write_MFRC522(reg, tmp | mask)
    
  def ClearBitMask(self, reg, mask):
    tmp = self.Read_MFRC522(reg);
    self.Write_MFRC522(reg, tmp & (~mask))
  
  def AntennaOn(self):
  # turn antenna on if currently off
    temp = self.Read_MFRC522(self.TxControlReg)
    if(~(temp & 0x03)):
        self.SetBitMask(self.TxControlReg, 0x03)
  
  def AntennaOff(self):
    # turn off antenna regardless of previous state
    self.ClearBitMask(self.TxControlReg, 0x03)
	
  def FF32_Init(self):
    # Check for FF32 chip and configure SPI interface
    # Get silicon version
    with pyff32.FF32() as ff32:
        try:
            version = ff32.getChipInfo()
            if (version):
                print("FF32_Init: Chip is FF32, silicon version %d.%d" % version)
            else:
                print("FF32_Init: Chip is not FF32")
        except Exception as ex: 
            print("FF32_Init: *** Error (getChipInfo): " + ex.message)
        # Get chip address
        try:
            addr = ff32.getAddress()
            print ("FF32_Init: Chip address is %d" % (addr))
        except Exception as ex: 
            print("FF32_Init: *** Error (getAddress): " + ex.message)
        # Get board vendor
        try:
            vendor = ff32.getVendor()
            print("FF32_Init: Board vendor is %s" % (vendor))
        except Exception as ex: 
            print("FF32_Init: *** Error (getVendor): " + ex.message)
        # Get board name
        try:
            product = ff32.getProduct()
            print("FF32_Init: Board name is %s" % (product))
        except Exception as ex: 
            print("FF32_Init: *** Error (getProduct): " + ex.message)
        # Get board serial number
        try:
            serial = ff32.getSerialNumber()
            print("FF32_Init: Board serial number is %s" % (serial))
        except Exception as ex: 
            print("FF32_Init: *** Error (getSerialNumber): " + ex.message)
        # Set SPI bus pins
        try:
            ff32.setSPIPins(("A", 1), ("A", 2), ("A", 3), ("A", 4))
            print("FF32_Init: SPI bus pins have been set to SS=A1,SCK=A2, MOSI=A3, MISO=A4")
        except Exception as ex:
            print("FF32_Init: *** Error (SetSPIPins): " + ex.message)
	  
  def MFRC522_Init(self):
    # Once SPI communication is established program MFRC522 registers with initial values
    self.MFRC522_Reset();                                       # Reset MFRD522
    if self.Read_MFRC522(self.VersionReg)==0x91:                # Read and check RC522 version number
        print("MFRC522_Init: RC522 version 1.0 connected")
    else:
        if self.Read_MFRC522(self.VersionReg)==0x92:
            print("MFRC522_Init: RC522 version 2.0 connected")
        else:
            print("MFRC522_Init: No RC522 detected!")
            return
    self.Write_MFRC522(self.TModeReg, 0x8D)						# Program timer to 2kHz
    self.Write_MFRC522(self.TPrescalerReg, 0x3E)				#
    self.Write_MFRC522(self.TReloadRegL, 30)					# timer reload value 15,5 ms
    self.Write_MFRC522(self.TReloadRegH, 0)						#
    self.Write_MFRC522(self.TxASKReg, 0x40)						# Use 100% ASK modulation
    self.Write_MFRC522(self.ModeReg, 0x3D)						# MHIN active Hi, CRC preset 0x6363
    self.AntennaOn()											# Switch on Antenna

  def MFRC522_Selftest(self):
    selfTestResultV1 = bytearray([0x00, 0xC6, 0x37, 0xD5, 0x32, 0xB7, 0x57, 0x5C,
                          0xC2, 0xD8, 0x7C, 0x4D, 0xD9, 0x70, 0xC7, 0x73,
                          0x10, 0xE6, 0xD2, 0xAA, 0x5E, 0xA1, 0x3E, 0x5A,
                          0x14, 0xAF, 0x30, 0x61, 0xC9, 0x70, 0xDB, 0x2E,
                          0x64, 0x22, 0x72, 0xB5, 0xBD, 0x65, 0xF4, 0xEC,
                          0x22, 0xBC, 0xD3, 0x72, 0x35, 0xCD, 0xAA, 0x41,
                          0x1F, 0xA7, 0xF3, 0x53, 0x14, 0xDE, 0x7E, 0x02,
                          0xD9, 0x0F, 0xB5, 0x5E, 0x25, 0x1D, 0x29, 0x79])
    selfTestResultV2 = bytearray([0x00, 0xEB, 0x66, 0xBA, 0x57, 0xBF, 0x23, 0x95,
                          0xD0, 0xE3, 0x0D, 0x3D, 0x27, 0x89, 0x5C, 0xDE,
                          0x9D, 0x3B, 0xA7, 0x00, 0x21, 0x5B, 0x89, 0x82,
                          0x51, 0x3A, 0xEB, 0x02, 0x0C, 0xA5, 0x00, 0x49,
                          0x7C, 0x84, 0x4D, 0xB3, 0xCC, 0xD2, 0x1B, 0x81,
                          0x5D, 0x48, 0x76, 0xD5, 0x71, 0x61, 0x21, 0xA9,
                          0x86, 0x96, 0x83, 0x38, 0xCF, 0x9D, 0x5B, 0x6D,
                          0xDC, 0x15, 0xBA, 0x3E, 0x7D, 0x95, 0x3B, 0x2F])
    if self.Read_MFRC522(self.VersionReg)==0x91:                # Read and check RC522 version number
        selfTestResult = selfTestResultV1
    else:
        if self.Read_MFRC522(self.VersionReg)==0x92:
            selfTestResult = selfTestResultV2
        else:
            return false
    self.MFRC522_Reset()
    self.Write_MFRC522(self.FIFODataReg, 0x00)
    self.Write_MFRC522(self.CommandReg, self.PCD_MEM)
    self.Write_MFRC522(self.AutoTestReg, 0x09)           # enable self test
    self.Write_MFRC522(self.FIFODataReg, 0x00)
    self.Write_MFRC522(self.CommandReg, self.PCD_CALC_CRC);   # start self test

    # Wait for the self test to complete.
    i = 0xFF
    while ((i != 0) & (0x04 & self.Read_MFRC522(self.DivIrqReg))):
        i -= 1
  
    # read 64 byte from FIFO and compare with data expected
    for i in range (0,64):
        if (self.Read_MFRC522(self.FIFODataReg) != selfTestResult[i]):
            print("MFRC522_Selftest: Test failed at byte {0:02}".format(i))
            return False

    return True

  def MFRC522_ToCard(self,command,sendData):
    backData = []
    backLen = 0
    status = self.MI_ERR
    irqEn = 0x00
    waitIRq = 0x00
    lastBits = None
    n = 0
    i = 0
    
    if command == self.PCD_AUTHENT:
        irqEn = 0x12
        waitIRq = 0x10
    if command == self.PCD_TRANSCEIVE:
        irqEn = 0x77
        waitIRq = 0x30
    
    self.Write_MFRC522(self.CommIEnReg, irqEn|0x80)
    self.ClearBitMask(self.CommIrqReg, 0x80)
    self.SetBitMask(self.FIFOLevelReg, 0x80)
    
    self.Write_MFRC522(self.CommandReg, self.PCD_IDLE) 
    
    # write sendData to FIFO
    while(i<len(sendData)):
        self.Write_MFRC522(self.FIFODataReg, sendData[i])
        i += 1
    
    self.Write_MFRC522(self.CommandReg, command)
    
    # start transmission of data  
    if command == self.PCD_TRANSCEIVE:
        self.SetBitMask(self.BitFramingReg, 0x80)
    
    i = 10
    while True:
        n = self.Read_MFRC522(self.CommIrqReg)
        time.sleep(0.1)
        i -= 1
        if ~((i!=0) and ~(n&0x01) and ~(n&waitIRq)):
            break
            
    # stop transmission of data
    self.ClearBitMask(self.BitFramingReg, 0x80)
    
    # was a card detected before countdown reached then read FIFO
    if i != 0:
        if (self.Read_MFRC522(self.ErrorReg) & 0x1B)==0x00:
            status = self.MI_OK

            if n & irqEn & 0x01:
                status = self.MI_NOTAGERR
      
            if command == self.PCD_TRANSCEIVE:
                n = self.Read_MFRC522(self.FIFOLevelReg)
                lastBits = self.Read_MFRC522(self.ControlReg) & 0x07
                if lastBits != 0:
                    backLen = (n-1)*8 + lastBits
                else:
                    backLen = n*8
          
            if n == 0:
                n = 1
            if n > self.MAX_LEN:
                n = self.MAX_LEN
    
            i = 0
            while i<n:
                backData.append(self.Read_MFRC522(self.FIFODataReg))
                i += 1
        else:
            status = self.MI_ERR

    return (status,backData,backLen)

  def MFRC522_Request(self, reqMode):
    status = None
    backBits = None
    TagType = []
    
    # TxLastBits set to 0b111: 
    self.Write_MFRC522(self.BitFramingReg, 0x07)
    
    TagType.append(reqMode);
    (status,backData,backBits) = self.MFRC522_ToCard(self.PCD_TRANSCEIVE, TagType)
  
    if ((status != self.MI_OK) | (backBits != 0x10)):
      status = self.MI_ERR
      
    return (status,backBits)
    
  def MFRC522_Anticoll(self):
    backData = []
    serNumCheck = 0
    serNum = []
  
    self.Write_MFRC522(self.BitFramingReg, 0x00)
    
    serNum.append(self.PICC_ANTICOLL)
    serNum.append(0x20)
    
    (status,backData,backBits) = self.MFRC522_ToCard(self.PCD_TRANSCEIVE,serNum)
    
    if(status == self.MI_OK):
        i = 0
        if len(backData)==5:
            while i<4:
                serNumCheck = serNumCheck ^ backData[i]
                i += 1
            if serNumCheck != backData[i]:
                status = self.MI_ERR
        else:
            status = self.MI_ERR
  
    return (status,backData)  

 
  def CalulateCRC(self, pIndata):
    self.ClearBitMask(self.DivIrqReg, 0x04)
    self.SetBitMask(self.FIFOLevelReg, 0x80);
    i = 0
    while i<len(pIndata):
        self.Write_MFRC522(self.FIFODataReg, pIndata[i])
        i += 1
    self.Write_MFRC522(self.CommandReg, self.PCD_CALC_CRC)
    i = 0xFF
    while True:
        n = self.Read_MFRC522(self.DivIrqReg)
        i -= 1
        if not ((i != 0) and not (n&0x04)):
            break
    pOutData = []
    pOutData.append(self.Read_MFRC522(self.CRCResultRegL))
    pOutData.append(self.Read_MFRC522(self.CRCResultRegM))
    return pOutData
  
  def MFRC522_SelectTag(self, serNum):
    backData = []
    buf = []
    buf.append(self.PICC_SElECTTAG)
    buf.append(0x70)
    i = 0
    while i<5:
        buf.append(serNum[i])
        i += 1
    pOut = self.CalulateCRC(buf)
    buf.append(pOut[0])
    buf.append(pOut[1])
    (status, backData, backLen) = self.MFRC522_ToCard(self.PCD_TRANSCEIVE, buf)
    
    if (status == self.MI_OK) and (backLen == 0x18):
        print("MFRC522_SelectTag: Size: " + str(backData[0]))
        return  backData[0]
    else:
        return 0
  
  def MFRC522_Auth(self, authMode, BlockAddr, Sectorkey, serNum):
    buff = []
    buff.append(authMode)
    buff.append(BlockAddr)
    i = 0
    while(i < len(Sectorkey)):
        buff.append(Sectorkey[i])
        i += 1
    i = 0
    while(i < len(serNum)):
        buff.append(serNum[i])
        i += 1
    (status, backData, backLen) = self.MFRC522_ToCard(self.PCD_AUTHENT,buff)
    if not(status == self.MI_OK):
        print("MFRC522_Auth: AUTH ERROR! Status returned ",status)
    if not (self.Read_MFRC522(self.Status2Reg) & 0x08) != 0:
        print("MFRC522_Auth: AUTH ERROR(status2reg & 0x08) != 0")

    return status 

 
#---------------- start of main program -------------------#  
do_selftest = False
print("Main: Initializing...")
MIFAREReader = MFRC522()
# Note that selftest disrupts some registers. A MFRC522_Init() is needed to get things working again after a selftest...
if do_selftest:
    if MIFAREReader.MFRC522_Selftest():
        print("Main: RC522 Selftest passed...")
    else:
        print("Main: RC522 Selftest failed...")  

while True:
    (status,TagType) = MIFAREReader.MFRC522_Request(MIFAREReader.PICC_REQIDL)
    if status == MIFAREReader.MI_OK:
        print("Main: Card detected")
        (status,backData) = MIFAREReader.MFRC522_Anticoll()
        if status == MIFAREReader.MI_OK:
            print("Main: Card read UID: "+str(backData[0])+","+str(backData[1])+","+str(backData[2])+","+str(backData[3])+","+str(backData[4]))
            if backData==([100,25,206,171,24]):
                with pyff32.FF32() as ff32:
                    try:
                        ff32.setDigitalOutput(("B", 1), 0)
                    except Exception as ex: 
                        print("*** Error (setDigitalOutput): " + ex.message)
                    time.sleep(0.3)
                    try:
                        ff32.setDigitalOutput(("B", 1), 1)
                    except Exception as ex: 
                        print("*** Error (setDigitalOutput): " + ex.message)
            else:
                 with pyff32.FF32() as ff32:
                    try:
                        ff32.setDigitalOutput(("B", 1), 0)
                    except Exception as ex: 
                        print("*** Error (setDigitalOutput): " + ex.message)
                    time.sleep(0.6)
                    try:
                        ff32.setDigitalOutput(("B", 1), 1)
                    except Exception as ex: 
                        print("*** Error (setDigitalOutput): " + ex.message)
                        


  

 
 